﻿#region License
/*
Kangaroo Library for C# and .NET
Copyright (c) 2013-2014 Dimension Engineering LLC
http://www.dimensionengineering.com/kangaroo

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
USE OR PERFORMANCE OF THIS SOFTWARE.
*/
#endregion

using System;
using System.IO;
using System.Threading;

namespace DimensionEngineering.Kangaroo
{
    /// <summary>
    /// Lets you send commands to and get position, speed, etc. for a specific channel of the Kangaroo.
    /// </summary>
    public class KangarooChannel
    {
        /// <summary>
        /// Constructs a KangarooChannel object.
        /// </summary>
        /// <param name="serial">The <see cref="KangarooSerial"/> whose serial port the controller is on.</param>
        /// <param name="name">
        /// The name of the channel (a single character).
        /// By default, mixed mode channels are 'D' (Drive) and 'T' (Turn). Independent mode channels are '1' and '2'.
        /// You can change these in DEScribe. Mainly this is useful if you have multiple Kangaroos daisy-chained and have
        /// 'Enable multi-Kangaroo mode (shared signal lines).' checked, or if your particular application would be clearer
        /// if for instance the channels were named 'X', 'Y', and 'Z'.
        /// </param>
        /// <param name="address">
        /// The Packet Serial address of the controller.
        /// Normally this can be left at its default. If you daisy-chain the TX/S1 line between your Kangaroo and a 
        /// Sabertooth or SyRen motor driver set in Packet Serial, however, make sure they aren't both using the same address.
        /// </param>
        public KangarooChannel(KangarooSerial serial, char name, int address = KangarooConstants.DefaultAddress)
        {
            if (serial == null) { throw new ArgumentNullException("serial"); }
            if ((ushort)name >= 128) { throw new ArgumentOutOfRangeException("name"); }
            if (address < 128 || address > 255) { throw new ArgumentOutOfRangeException("address"); }
            _kangaroo = serial; _address = (byte)address;

            if (name >= 'a' && name <= 'z')
            {
                name = (char)(name - 'a' + 'A');
            }

            if ((name >= 'A' && name <= 'Z') ||
                (name >= '0' && name <= '9'))
            {
                _name = (byte)name;
            }

            CommandRetryInterval = KangarooConstants.DefaultCommandRetryInterval;
            CommandTimeout = KangarooConstants.DefaultCommandTimeout;
        }

        /// <summary>
        /// Starts the channel. Also, the Kangaroo LED will shine brightly for a third of a second.
        /// </summary>
        /// <returns>
        /// A <see cref="KangarooError"/>.
        /// Most commonly, this will be <see cref="KangarooError.None"/> if the channel does not require homing,
        /// or <see cref="KangarooError.NotHomed"/> if it does.
        /// </returns>
        public KangarooError Start()
        {
            KangarooCommandWriter contents = new KangarooCommandWriter();
            return Set(KangarooCommand.Start, contents).Status.Error;
        }

        /// <summary>
        /// Sets custom units for the channel.
        /// This command may be called after you start the channel but before you home it.
        /// If you do not set custom units, the units you have set up in DEScribe will be used.
        /// If you haven't set any, machine units will be used.
        /// </summary>
        /// <param name="desiredUnits">The amount in your units that correspond to the specified amount of machine units.</param>
        /// <param name="machineUnits">The amount of machine units (millivolts or lines) corresponding to the specified amount in your units.</param>
        /// <returns>
        /// A <see cref="KangarooError"/>.
        /// Most commonly, this will be <see cref="KangarooError.None"/> if the channel does not require homing,
        /// or <see cref="KangarooError.NotHomed"/> if it does.
        /// </returns>
        public KangarooError Units(int desiredUnits, int machineUnits)
        {
            if (desiredUnits <= 0) { throw new ArgumentOutOfRangeException("desiredUnits"); }
            if (machineUnits <= 0) { throw new ArgumentOutOfRangeException("machineUnits"); }

            KangarooCommandWriter contents = new KangarooCommandWriter();
            contents.WriteBitPackedNumber(desiredUnits);
            contents.WriteBitPackedNumber(machineUnits);
            return Set(KangarooCommand.Units, contents).Status.Error;
        }

        /// <summary>
        /// Homes the channel.
        /// </summary>
        /// <returns>
        /// A <see cref="KangarooMonitor"/> for tracking the homing request.
        /// The easiest way to use the KangarooMonitor here is to call <see cref="KangarooMonitor.Wait(int)"/>.
        /// </returns>
        public KangarooMonitor Home()
        {
            KangarooCommandWriter contents = new KangarooCommandWriter();
            return Set(KangarooCommand.Home, contents);
        }

        /// <summary>
        /// Moves to the specified absolute position.
        /// This command is most useful for absolutely-positioned systems.
        /// </summary>
        /// <param name="position">The position to move to.</param>
        /// <param name="speedLimit">The speed limit for the move.</param>
        /// <param name="flags">Modifiers for the move command.</param>
        /// <returns>A <see cref="KangarooMonitor"/> for tracking the move request.</returns>
        public KangarooMonitor P(int position,
                                 int speedLimit = KangarooConstants.UnspecifiedLimit,
                                 KangarooMoveFlags flags = KangarooMoveFlags.Default)
        {
            return Motion((byte)0x01, position, (byte)0x02, speedLimit, flags);
        }

        /// <summary>
        /// Makes an incremental move.
        /// This command is useful any time you want to make a motion relative to where you are right now.
        /// Rovers are a case where this is almost always what you want.
        /// </summary>
        /// <param name="positionIncrement">The amount to increment the current position by.</param>
        /// <param name="speedLimit">The speed limit for the move.</param>
        /// <param name="flags">Modifiers for the move command.</param>
        /// <returns>A <see cref="KangarooMonitor"/> for tracking the move request.</returns>
        public KangarooMonitor PI(int positionIncrement,
                                  int speedLimit = KangarooConstants.UnspecifiedLimit,
                                  KangarooMoveFlags flags = KangarooMoveFlags.Default)
        {
            return Motion((byte)0x41, positionIncrement, (byte)0x02, speedLimit, flags);
        }

        /// <summary>
        /// Moves at a particular speed.
        /// </summary>
        /// <param name="velocity">The speed to move at.</param>
        /// <param name="rampLimit">The max rate at which to change the commanded speed.</param>
        /// <param name="flags">Modifiers for the move command.</param>
        /// <returns>A <see cref="KangarooMonitor"/> for tracking the move request.</returns>
        public KangarooMonitor S(int velocity,
                                 int rampLimit = KangarooConstants.UnspecifiedLimit,
                                 KangarooMoveFlags flags = KangarooMoveFlags.Default)
        {
            return Motion((byte)0x02, velocity, (byte)0x03, rampLimit, flags);
        }

        /// <summary>
        /// Moves at a particular speed, incremental from the current speed.
        /// This is useful for cruise control: a negative increment will slow down,
        /// a positive increment will speed up, and a zero increment will hold the current speed.
        /// </summary>
        /// <param name="velocityIncrement">The amount to increment the current speed by.</param>
        /// <param name="rampLimit">The max rate at which to change the commanded speed.</param>
        /// <param name="flags">Modifiers for the move command.</param>
        /// <returns>A <see cref="KangarooMonitor"/> for tracking the move request.</returns>
        public KangarooMonitor SI(int velocityIncrement,
                                  int rampLimit = KangarooConstants.UnspecifiedLimit,
                                  KangarooMoveFlags flags = KangarooMoveFlags.Default)
        {
            return Motion((byte)0x42, velocityIncrement, (byte)0x03, rampLimit, flags);
        }

        /// <summary>
        /// Issues a 'get' request.
        /// </summary>
        /// <param name="type">The type of the 'get' request.</param>
        /// <param name="flags">Flags modifying the 'get' request.</param>
        /// <returns>A <see cref="KangarooStatus"/> object describing the response.</returns>
        public KangarooStatus Get(KangarooGetType type, KangarooGetFlags flags = KangarooGetFlags.Default)
        {
            KangarooTimeout timeout = new KangarooTimeout(CommandTimeout);

            KangarooStatus initialStatus = GetInitialSequenceCodeIfNecessary(timeout);
            if (initialStatus != null) { return initialStatus; }

            return GetSpecial(type, flags, timeout);
        }

        /// <summary>
        /// Gets the absolute position.
        /// </summary>
        /// <param name="flags">Flags modifying the 'get' request.</param>
        /// <returns>A <see cref="KangarooStatus"/> object describing the position.</returns>
        public KangarooStatus GetP(KangarooGetFlags flags = KangarooGetFlags.Default)
        {
            return Get(KangarooGetType.P, flags);
        }

        /// <summary>
        /// Gets the incremental position (relative to the position when the last command was issued).
        /// </summary>
        /// <param name="flags">Flags modifying the 'get' request.</param>
        /// <returns>A <see cref="KangarooStatus"/> object describing the position.</returns>
        public KangarooStatus GetPI(KangarooGetFlags flags = KangarooGetFlags.Default)
        {
            return Get(KangarooGetType.PI, flags);
        }

        /// <summary>
        /// Gets the absolute speed (positive or negative depending on direction).
        /// </summary>
        /// <param name="flags">Flags modifying the 'get' request.</param>
        /// <returns>A <see cref="KangarooStatus"/> object describing the speed.</returns>
        public KangarooStatus GetS(KangarooGetFlags flags = KangarooGetFlags.Default)
        {
            return Get(KangarooGetType.S, flags);
        }

        /// <summary>
        /// Gets the incremental speed (relative to the speed when the last command was issued).
        /// </summary>
        /// <param name="flags">Flags modifying the 'get' request.</param>
        /// <returns>A <see cref="KangarooStatus"/> object describing the speed.</returns>
        public KangarooStatus GetSI(KangarooGetFlags flags = KangarooGetFlags.Default)
        {
            return Get(KangarooGetType.SI, flags);
        }

        /// <summary>
        /// Gets the minimum position. This corresponds to DEScribe's Nominal Travel minimum.
        /// </summary>
        /// <param name="flags">Flags modifying the 'get' request.</param>
        /// <returns>A <see cref="KangarooStatus"/> object describing the limit.</returns>
        public KangarooStatus GetMin(KangarooGetFlags flags = KangarooGetFlags.Default)
        {
            return Get(KangarooGetType.Min, flags);
        }

        /// <summary>
        /// Gets the maximum position. This corresponds to DEScribe's Nominal Travel maximum.
        /// </summary>
        /// <param name="flags">Flags modifying the 'get' request.</param>
        /// <returns>A <see cref="KangarooStatus"/> object describing the limit.</returns>
        public KangarooStatus GetMax(KangarooGetFlags flags = KangarooGetFlags.Default)
        {
            return Get(KangarooGetType.Max, flags);
        }

        /// <summary>
        /// Sets the baud rate.
        /// This affects all channels of the controller that receives the command.
        /// </summary>
        /// <param name="baudRate">The baud rate to use. Supported rates are 9600, 19200, 38400, and 115200.</param>
        public void BaudRate(int baudRate)
        {
            int index;
            switch (baudRate)
            {
                case 9600:   index = 0; break;
                case 19200:  index = 1; break;
                case 38400:  index = 2; break;
                case 115200: index = 3; break;
                default: throw new ArgumentOutOfRangeException("baudRate");
            }

            SystemCommand(KangarooSystemCommand.SetBaudRate, false, index);
        }

        /// <summary>
        /// Powers down the channel.
        /// </summary>
        /// <returns>
        /// A <see cref="KangarooError"/>, usually <see cref="KangarooError.None"/>.
        /// </returns>
        public KangarooError PowerDown()
        {
            return SystemCommand(KangarooSystemCommand.PowerDown, true);
        }

        /// <summary>
        /// Powers down all channels of the controller that receives the command.
        /// </summary>
        /// <returns>
        /// A <see cref="KangarooError"/>, usually <see cref="KangarooError.None"/>.
        /// </returns>
        public KangarooError PowerDownAll()
        {
            return SystemCommand(KangarooSystemCommand.PowerDownAll, true);
        }

        /// <summary>
        /// Sets the serial timeout.
        /// This affects all channels of the controller that receives the command.
        /// </summary>
        /// <param name="milliseconds">The timeout, in milliseconds. A value of zero uses the DEScribe setting. <see cref="Timeout.Infinite"/> disables the timeout.</param>
        /// <returns>
        /// A <see cref="KangarooError"/>, usually <see cref="KangarooError.None"/>.
        /// </returns>
        public KangarooError SerialTimeout(int milliseconds)
        {
            if (milliseconds < 0)
            {
                if (milliseconds != Timeout.Infinite)
                {
                    throw new ArgumentOutOfRangeException("milliseconds",
                                                          "Timeout is negative and not equal to System.Threading.Timeout.Infinite.");
                }
            }

            return SystemCommand(KangarooSystemCommand.SetSerialTimeout, true,
                                 milliseconds < 0 ? -1 : (milliseconds * 2 + 124) / 125); // The command takes time in 16ths of a second. Round up.
        }

        KangarooStatus GetSpecial(KangarooGetType type, KangarooGetFlags flags, KangarooTimeout timeout)
        {
            KangarooTimeout retry = new KangarooTimeout(CommandRetryInterval); retry.Expire();
            flags = (KangarooGetFlags)(flags | KangarooGetFlags.EchoCode);
  
            while (true)
            {
                if (!_kangaroo.IsOpen)
                {
                    return KangarooStatus.CreatePortNotOpen();
                }

                if (timeout.Expired)
                {
                    return KangarooStatus.CreateTimedOut();
                }

                if (retry.Expired)
                {
                    retry.Reset();
                    _echoCode = NextCode(_echoCode);
      
                    // Send the Get command.
                    KangarooCommandWriter writer = new KangarooCommandWriter();
                    writer.Write((byte)Name);
                    writer.Write((byte)flags);
                    writer.Write(_echoCode);
                    writer.Write((byte)type);

                    try
                    {
                        Stream port = _kangaroo.Port;
                        writer.WriteToStream(port, Address, KangarooCommand.Status);
                    }
                    catch (Exception)
                    {
                        _kangaroo.Close(); continue;
                    }
                }
    
                // Check for a valid reply.
                if (!_kangaroo.TryReceivePacket(timeout < retry ? timeout : retry)) { continue; }

                KangarooReplyReceiver receiver = _kangaroo._receiver;
                if (receiver.Address != Address) { continue; }
                if (receiver.Command != KangarooReplyCode.Status) { continue; }
    
                KangarooStatus result = new KangarooStatus(receiver.Buffer, receiver.Offset, receiver.Length);
                if (!result.IsValid) { continue; }

                if (result.EchoCode != _echoCode) { continue; }

                return result;
            }
        }

        KangarooMonitor Motion(byte motionType, int motionValue, byte limitType, int limitValue, KangarooMoveFlags flags)
        {
            KangarooCommandWriter contents = new KangarooCommandWriter();

            contents.Write(motionType);
            contents.WriteBitPackedNumber(motionValue);

            if (limitValue >= 0)
            {
                contents.Write(limitType);
                contents.WriteBitPackedNumber(limitValue);
            }
            else if (limitValue != KangarooConstants.UnspecifiedLimit)
            {
                throw new ArgumentOutOfRangeException("Limit is negative and not equal to Kangaroo.UnspecifiedLimit.",
                                                      (Exception)null);
            }

            return Set(KangarooCommand.Move, contents, flags);
        }

        /// <summary>
        /// Sends a system command. These are mostly for advanced use cases.
        /// </summary>
        /// <param name="systemCommand">The command to set.</param>
        /// <param name="expectReply">
        /// Whether or not to expect a reply.
        /// If a command has an unusual effect on the state of the Kangaroo, expecting a reply may
        /// cause the command to retry until it times out.
        /// </param>
        /// <param name="values">The parameters of the command.</param>
        /// <returns>A <see cref="KangarooError"/>.</returns>
        public KangarooError SystemCommand(KangarooSystemCommand systemCommand, bool expectReply, params int[] values)
        {
            if (values == null) { throw new ArgumentNullException("values"); }

            KangarooCommandWriter contents = new KangarooCommandWriter();
            contents.Write((byte)systemCommand);
            foreach (int value in values) { contents.WriteBitPackedNumber(value); }

            if (expectReply)
            {
                return Set(KangarooCommand.System, contents).Status.Error;
            }
            else
            {
                SetNoReply(KangarooCommand.System, contents);
                return KangarooError.None;
            }
        }

        void SetNoReply(KangarooCommand command, KangarooCommandWriter contents,
                        KangarooMoveFlags moveFlags = KangarooMoveFlags.Default)
        {
            Stream port = _kangaroo.Port;

            KangarooCommandWriter writer = new KangarooCommandWriter();
            writer.Write((byte)Name);
            writer.Write((byte)moveFlags);
            if (0 != (moveFlags & KangarooMoveFlags.SequenceCode))
            {
                writer.Write(_monitoredSequenceCode);
            }
            writer.Write(contents.Buffer, contents.Offset, contents.Length);

            try
            {
                writer.WriteToStream(port, Address, command);
            }
            catch (Exception)
            {
                _kangaroo.Close();
            }
        }

        // We return a status only if there is an error.
        KangarooStatus GetInitialSequenceCodeIfNecessary(KangarooTimeout timeout)
        {
            if (_monitoredSequenceCodeIsReady) { return null; }
            KangarooStatus status = GetSpecial(KangarooGetType.P, KangarooGetFlags.SequenceCode, timeout);
            if ((int)status.Error < 0) { return status; }

            _monitoredSequenceCode = status.SequenceCode;
            _monitoredSequenceCodeIsReady = true;
            return null;
        }

        KangarooMonitor Set(KangarooCommand command, KangarooCommandWriter contents,
                            KangarooMoveFlags moveFlags = KangarooMoveFlags.Default,
                            KangarooGetType getType = KangarooGetType.P)
        {
            KangarooMonitor monitor = new KangarooMonitor(this);

            if (Streaming)
            {
                SetNoReply(command, contents, moveFlags);

                // Motions from other commands cannot be tracked through.
                _monitor = null;
            }
            else
            {
                KangarooTimeout timeout = new KangarooTimeout(CommandTimeout);
                moveFlags |= KangarooMoveFlags.SequenceCode;

                // Invalidate the older command monitor and set up a new one.
                _monitor = monitor;
                _monitoredGetResult = GetInitialSequenceCodeIfNecessary(timeout);

                // If we have a proper initial sequence code, send the command.
                if (_monitoredGetResult == null)
                {
                    _monitoredGetType = getType;
                    _monitoredGetFlags = (KangarooGetFlags)moveFlags;
                    _monitoredSequenceCode = NextCode(_monitoredSequenceCode);

                    do
                    {
                        SetNoReply(command, contents, moveFlags);
                    }
                    // Make sure the packet was received correctly -- proper sequence code and the like. If not, we'll resend.
                    while (!UpdateMonitoredResult(timeout, false));
                }
            }
  
            return monitor;
        }

        internal bool UpdateMonitoredResult(KangarooTimeout timeout, bool acceptRepliesWithStartupSequenceCode)
        {
            _monitoredGetResult = GetSpecial(_monitoredGetType, _monitoredGetFlags, timeout);

            if ((int)_monitoredGetResult.Error < 0)
            {
                return true;
            }

            if (_monitoredGetResult.SequenceCode == _monitoredSequenceCode)
            {
                return true;
            }

            if (_monitoredGetResult.SequenceCode == 0 && acceptRepliesWithStartupSequenceCode)
            {
                return true;
            }

            return false;
        }

        static byte NextCode(byte code)
        {
            if (++code >= 0x80) { code = 1; } // The device will start up with code 0, so we do this to distinguish that case.
            return code;
        }

        /// <summary>
        /// Gets or sets the command retry interval, in milliseconds.
        /// </summary>
        public int CommandRetryInterval
        {
            get { return _commandRetryInterval; }
            set
            {
                if (value < 0)
                {
                    throw new ArgumentOutOfRangeException("Retry interval is negative or equal to System.Threading.Timeout.Infinite.",
                                                          (Exception)null);
                }

                _commandRetryInterval = value;
            }
        }

        /// <summary>
        /// Gets or sets the command timeout, in milliseconds.
        /// </summary>
        public int CommandTimeout
        {
            get { return _commandTimeout; }
            set
            {
                if (value < 0)
                {
                    if (value != Timeout.Infinite)
                    {
                        throw new ArgumentOutOfRangeException("Timeout is negative and not equal to System.Threading.Timeout.Infinite.",
                                                              (Exception)null);
                    }
                }

                _commandTimeout = value;
            }
        }

        /// <summary>
        /// Enables or disables streaming.
        /// When streaming is disabled (the default), commands are delivered reliably:
        /// a request is sent to the Kangaroo after every command to make sure the command was received.
        /// When streaming is enabled, commands are delivered unreliably:
        /// no effort is made to verify that they are received.
        ///
        /// If you are constantly sending position or speed commands, you may not care if an intermediate one is lost.
        /// Enabling streaming can more than double your command rate, at the cost of reduced reliability.
        /// If you only have a transmit line and no way to receive data back from the Kangaroo, you should enable streaming.
        /// <see cref="KangarooMonitor.Wait(int)"/> cannot be meaningfully called for a command that was streamed.
        /// </summary>
        public bool Streaming
        {
            get { return _streaming; }
            set { _streaming = value; }
        }

        /// <summary>
        /// Gets the Packet Serial address of the controller.
        /// </summary>
        public int Address
        {
            get { return _address; }
        }

        /// <summary>
        /// Gets the channel name.
        /// </summary>
        public char Name
        {
            get { return (char)_name; }
        }

        KangarooSerial           _kangaroo;
        byte                     _name;
        byte                     _address;
        int                      _commandRetryInterval;
        int                      _commandTimeout;
        byte                     _echoCode;
        internal KangarooMonitor _monitor;
        KangarooGetType          _monitoredGetType;
        KangarooGetFlags         _monitoredGetFlags;
        internal KangarooStatus  _monitoredGetResult;
        byte                     _monitoredSequenceCode;
        bool                     _monitoredSequenceCodeIsReady;
        bool                     _streaming;
    }
}
