﻿#region License
/*
Kangaroo Library for C# and .NET
Copyright (c) 2013 Dimension Engineering LLC
http://www.dimensionengineering.com/kangaroo

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
USE OR PERFORMANCE OF THIS SOFTWARE.
*/
#endregion

using System;
using System.Threading;

namespace DimensionEngineering.Kangaroo
{
    /// <summary>
    /// Makes it easy to track the status of position commands, speed commands, and any other commands that take time to complete.
    /// </summary>
    public class KangarooMonitor
    {
        KangarooChannel _channel;

        internal KangarooMonitor(KangarooChannel channel)
        {
            _channel = channel;
        }

        /// <summary>
        /// Updates the status. The channel's command timeout is used.
        /// </summary>
        /// <returns>The same KangarooMonitor.</returns>
        public KangarooMonitor Update()
        {
            return Update(_channel.CommandTimeout);
        }

        /// <summary>
        /// Updates the status.
        /// </summary>
        /// <param name="timeoutMS">The timeout to use, in milliseconds.</param>
        /// <returns>The same KangarooMonitor.</returns>
        public KangarooMonitor Update(int timeoutMS)
        {
            return Update(new KangarooTimeout(timeoutMS));
        }

        /// <summary>
        /// Updates the status.
        /// </summary>
        /// <param name="timeout">The timeout object to use.</param>
        /// <returns>The same KangarooMonitor.</returns>
        public KangarooMonitor Update(KangarooTimeout timeout)
        {
            while (IsValid && !_channel.UpdateMonitoredResult(timeout, true)) ;
            return this;
        }

        /// <summary>
        /// Waits until <see cref="KangarooStatus.IsDone"/> returns true or the timeout expires.
        /// </summary>
        /// <param name="timeoutMS">The timeout to use, in milliseconds.</param>
        /// <returns>The same KangarooMonitor.</returns>
        public KangarooMonitor Wait(int timeoutMS = Timeout.Infinite)
        {
            return Wait(new KangarooTimeout(timeoutMS));
        }

        /// <summary>
        /// Waits until <see cref="KangarooStatus.IsDone"/> returns true or the timeout expires.
        /// </summary>
        /// <param name="timeout">The timeout object to use.</param>
        /// <returns>The same KangarooMonitor.</returns>
        public KangarooMonitor Wait(KangarooTimeout timeout)
        {
            while (!Status.IsDone) { Update(timeout); }
            return this;
        }

        /// <summary>
        /// Waits until all of the monitors are done or the timeout expires.
        /// </summary>
        /// <param name="monitors">The monitors to wait for.</param>
        /// <param name="timeoutMS">The timeout to use, in milliseconds.</param>
        /// <returns>True if all of the monitors are done, or false if the timeout expired.</returns>
        /// <remarks>If a monitor is null, it is ignored.</remarks>
        public static bool WaitAll(KangarooMonitor[] monitors, int timeoutMS = Timeout.Infinite)
        {
            return WaitAll(monitors, new KangarooTimeout(timeoutMS));
        }

        /// <summary>
        /// Waits until all of the monitors are done or the timeout expires.
        /// </summary>
        /// <param name="monitors">The monitors to wait for.</param>
        /// <param name="timeout">The timeout object to use.</param>
        /// <returns>True if all of the monitors are done, or false if the timeout expired.</returns>
        /// <remarks>If a monitor is null, it is ignored.</remarks>
        public static bool WaitAll(KangarooMonitor[] monitors, KangarooTimeout timeout)
        {
            if (monitors == null) { throw new ArgumentNullException("monitors"); }

            for (int i = 0; i < monitors.Length; i++)
            {
                KangarooMonitor monitor = monitors[i];
                if (monitor == null) { continue; }

                monitor.Wait(timeout);
                if (monitor.Status.TimedOut) { return false; }
            }

            return true;
        }

        /// <summary>
        /// Waits until at least one of the monitors is done or the timeout expires.
        /// </summary>
        /// <param name="monitors">The monitors to wait for.</param>
        /// <param name="timeoutMS">The timeout to use, in milliseconds.</param>
        /// <returns>The array index of the monitor that is done, or -1 if the timeout expired.</returns>
        /// <remarks>If a monitor is null, it is ignored.</remarks>
        public static int WaitAny(KangarooMonitor[] monitors, int timeoutMS)
        {
            return WaitAny(monitors, new KangarooTimeout(timeoutMS));
        }

        /// <summary>
        /// Waits until at least one of the monitors is done or the timeout expires.
        /// </summary>
        /// <param name="monitors">The monitors to wait for.</param>
        /// <param name="timeout">The timeout object to use.</param>
        /// <returns>The array index of the monitor that is done, or -1 if the timeout expired.</returns>
        /// <remarks>If a monitor is null, it is ignored.</remarks>
        public static int WaitAny(KangarooMonitor[] monitors, KangarooTimeout timeout)
        {
            if (monitors == null) { throw new ArgumentNullException("monitors"); }

            while (true)
            {
                if (timeout.Expired) { return -1; }

                for (int i = 0; i < monitors.Length; i++)
                {
                    KangarooMonitor monitor = monitors[i];
                    if (monitor == null) { continue; }

                    monitor.Update(timeout);
                    if (monitor.Status.TimedOut) { return -1; }
                    if (monitor.Status.IsDone) { return i; }
                }
            }
        }

        /// <summary>
        /// Gets whether the status is valid.
        /// If you keep an old KangarooMonitor object around and issue a new command on the same channel,
        /// the old KangarooMonitor will become invalid.
        /// </summary>
        public bool IsValid
        {
            get { return _channel._monitor == this; }
        }

        /// <summary>
        /// Gets the most recently received status.
        /// </summary>
        public KangarooStatus Status
        {
            get { return IsValid ? _channel._monitoredGetResult : KangarooStatus.CreateInvalidStatus(); }
        }
    }
}
