﻿#region License
/*
Kangaroo Library for C# and .NET
Copyright (c) 2013 Dimension Engineering LLC
http://www.dimensionengineering.com/kangaroo

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
USE OR PERFORMANCE OF THIS SOFTWARE.
*/
#endregion

using System;
using System.IO;
using System.IO.Ports;

namespace DimensionEngineering.Kangaroo
{
    /// <summary>
    /// Create a KangarooSerial for the serial port you are using,
    /// and then attach a <see cref="KangarooChannel"/> for each channel you want to communicate with.
    /// </summary>
    public class KangarooSerial : IDisposable
    {
        byte[] _buffer;
        int _bufferOffset;
        int _bufferLength;
        bool _open;
        Stream _port;
        IAsyncResult _read;
        internal KangarooReplyReceiver _receiver;

        /// <summary>
        /// Constructs a KangarooSerial object.
        /// </summary>
        public KangarooSerial()
        {
            _receiver = new KangarooReplyReceiver();
        }

        /// <summary>
        /// Opens the specified serial port.
        /// </summary>
        /// <param name="portName">The port name, such as COM1 or COM2.</param>
        /// <param name="baudRate">The baud rate.</param>
        public void Open(string portName, int baudRate = KangarooConstants.DefaultBaudRate)
        {
            if (portName == null) { throw new ArgumentNullException("portName"); }
            if (baudRate <= 0) { throw new ArgumentOutOfRangeException("baudRate"); }

            Close();

            SerialPort port = new SerialPort(portName, baudRate);
            try { port.Open(); Open(port); } catch { port.Close(); throw; }
        }

        /// <summary>
        /// Uses an already-open serial port.
        /// </summary>
        /// <param name="port">The serial port.</param>
        public void Open(SerialPort port)
        {
            if (port == null) { throw new ArgumentNullException("port"); }

            Open(new SerialStream(port));
        }

        /// <summary>
        /// Uses an already-open stream.
        /// </summary>
        /// <param name="port">The stream.</param>
        public void Open(Stream port)
        {
            if (port == null) { throw new ArgumentNullException("port"); }

            Close();

            _buffer = new byte[256];
            _bufferOffset = 0;
            _bufferLength = 0;
            _port = port;
            _read = null;
            _open = true;
        }

        /// <summary>
        /// Closes the serial port.
        /// </summary>
        public void Close()
        {
            if (!IsOpen) { return; }

            try
            {
                _port.Close();
            }
            catch (Exception)
            {

            }

            if (_read != null)
            {
                try
                {
                    _port.EndRead(_read);
                }
                catch (Exception)
                {

                }
            }

            _port = null;
            _read = null;
            _open = false;
        }

        internal bool TryReceivePacket(KangarooTimeout timeout)
        {
            while (true)
            {
                if (!IsOpen) { return false; }

                byte @byte;
                if (!TryReceiveByte(timeout, out @byte)) { return false; }
                
                _receiver.Read(@byte);
                if (_receiver.IsReady) { return true; }
            }
        }

        bool TryReceiveByte(KangarooTimeout timeout, out byte @byte)
        {
            @byte = 0;

            if (_bufferOffset < _bufferLength)
            {
                @byte = _buffer[_bufferOffset++];
                return true;
            }
            else
            {
                if (_read == null)
                {
                    try
                    {
                        _read = _port.BeginRead(_buffer, 0, _buffer.Length, null, null);
                    }
                    catch (Exception)
                    {
                        Close(); return false;
                    }
                }
                
                if (_read.AsyncWaitHandle.WaitOne(timeout.TimeLeft))
                {
                    int bytesRead;
                    try
                    {
                        bytesRead = _port.EndRead(_read);
                    }
                    catch (Exception)
                    {
                        _read = null;
                        Close(); return false;
                    }

                    _bufferOffset = 0;
                    _bufferLength = bytesRead;
                    _read = null;
                    if (bytesRead <= 0) { return false; }

                    @byte = _buffer[_bufferOffset++]; 
                    return true;
                }
                else
                {
                    return false;
                }
            }
        }

        void IDisposable.Dispose()
        {
            Close();
        }

        /// <summary>
        /// Gets whether the serial port is open.
        /// </summary>
        public bool IsOpen
        {
            get { return _open; }
        }

        /// <summary>
        /// Gets the serial port being used.
        /// </summary>
        public Stream Port
        {
            get
            {
                if (!IsOpen) { throw new InvalidOperationException("Port is not open."); }
                return _port;
            }
        }
    }
}
